/*
 *  Copyright 2018 The Kubernetes Authors.
 *  Copyright (c) 2019 NetApp
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *      http://www.apache.org/licenses/LICENSE-2.0
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package cloud

import (
	"context"
	"fmt"
	"path"
	"strings"
	"time"

	"cloud.google.com/go/compute/metadata"
	log "github.com/sirupsen/logrus"
	"golang.org/x/oauth2/google"
	compute "google.golang.org/api/compute/v0.beta"
	container "google.golang.org/api/container/v1"
	"google.golang.org/api/googleapi"
	"google.golang.org/api/option"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/util/wait"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

type InstanceStatus string

const (
	// Possible values
	Provisioning InstanceStatus = "PROVISIONING"
	Repairing    InstanceStatus = "REPAIRING"
	Running      InstanceStatus = "RUNNING"
	Staging      InstanceStatus = "STAGING"
	Stopped      InstanceStatus = "STOPPED"
	Stopping     InstanceStatus = "STOPPING"
	Suspended    InstanceStatus = "SUSPENDED"
	Suspending   InstanceStatus = "SUSPENDING"
	Terminated   InstanceStatus = "TERMINATED"
	Unknown      InstanceStatus = ""
)

// TODO:  Move all of these interfaces to cloud.go if they become used by other
// clouds.

// Data for an instance group as it pertains to a Kubernetes cluster. Note that
// MachineType isn't directly attainable through the compute.ManagedInstance,
// so we can't just use that.
type InstanceGroupClusterInfo interface {
	GetName() string
	GetZone() string
	GetMachineType() string
	GetNodePool() string
}

type InstanceGroup interface {
	GetName() string
	GetZone() string
	GetBaseInstanceName() string
	GetInstanceTemplate() string
	GetSize() int64
}

type ManagedInstance interface {
	GetName() string
	GetStatus() InstanceStatus
}

type gcpInstanceGroupClusterInfo struct {
	name        string
	zone        string
	machineType string
	nodePool    string
}

func (i *gcpInstanceGroupClusterInfo) GetName() string {
	return i.name
}
func (i *gcpInstanceGroupClusterInfo) GetZone() string {
	return i.zone
}
func (i *gcpInstanceGroupClusterInfo) GetMachineType() string {
	return i.machineType
}
func (i *gcpInstanceGroupClusterInfo) GetNodePool() string {
	return i.nodePool
}

type gcpManagedInstance struct {
	*compute.ManagedInstance
}

func (i *gcpManagedInstance) GetName() string {
	// Instance: [Output Only] The URL of the instance. The URL can exist
	// even if the instance has not yet been created.
	return path.Base(i.Instance)
}
func (i *gcpManagedInstance) GetStatus() InstanceStatus {
	return InstanceStatus(i.InstanceStatus)
}

type gcpInstanceGroup struct {
	*compute.InstanceGroupManager
}

func (i *gcpInstanceGroup) GetName() string {
	return i.Name
}
func (i *gcpInstanceGroup) GetZone() string {
	return path.Base(i.Zone)
}
func (i *gcpInstanceGroup) GetBaseInstanceName() string {
	return i.BaseInstanceName
}
func (i *gcpInstanceGroup) GetInstanceTemplate() string {
	return i.InstanceTemplate
}
func (i *gcpInstanceGroup) GetSize() int64 {
	return i.TargetSize
}

func ParseBaseInstancenameFromNodeName(nodeName string) string {
	return nodeName[0:strings.LastIndex(nodeName, "-")]
}
func ParseNodeGroupNameFromBaseInstanceName(baseName string) string {
	return fmt.Sprintf("%s-grp", baseName)
}
func ParseNodeGroupFromNodeName(nodeName string) string {
	return ParseNodeGroupNameFromBaseInstanceName(ParseBaseInstancenameFromNodeName(nodeName))
}

type GCPClient struct {
	project   string
	zone      string
	region    string
	cluster   string
	compute   *compute.Service
	container *container.Service
	sharedVIP bool
}

func isGCEError(err error, reason string) bool {
	apiErr, ok := err.(*googleapi.Error)
	if !ok {
		return false
	}

	for _, e := range apiErr.Errors {
		if e.Reason == reason {
			return true
		}
	}
	return false
}

func computeOpIsDone(op *compute.Operation) (bool, error) {
	if op == nil || op.Status != "DONE" {
		return false, nil
	}
	if op.Error != nil && len(op.Error.Errors) > 0 && op.Error.Errors[0] != nil {
		return true, fmt.Errorf("operation %v failed (%v): %v", op.Name, op.Error.Errors[0].Code, op.Error.Errors[0].Message)
	}
	return true, nil
}

func (client *GCPClient) waitForRegionalOp(ctx context.Context, op *compute.Operation) error {
	return wait.Poll(3*time.Second, 5*time.Minute, func() (bool, error) {
		pollOp, err := client.compute.RegionOperations.Get(client.project, client.region, op.Name).Context(ctx).Do()
		if err != nil {
			log.Errorf("WaitForOp(op: %#v, region: %#v) failed to poll the operation", op, client.region)
			return false, err
		}
		done, err := computeOpIsDone(pollOp)
		return done, err
	})
}

func (client *GCPClient) waitForZoneOp(
	ctx context.Context,
	zone string,
	op *compute.Operation,
	pollInterval time.Duration,
) error {
	err := wait.Poll(pollInterval, 5*time.Minute, func() (bool, error) {
		pollOp, err := client.compute.ZoneOperations.Get(client.project,
			zone, op.Name).Context(ctx).Do()
		if err != nil {
			log.Errorf("WaitForOp(op: %#v, zone: %#v) failed to poll the operation",
				op, client.region)
			return false, err
		}
		done, err := computeOpIsDone(pollOp)
		return done, err
	})
	if err != nil {
		log.Errorf("Operation error: %v\n", err)
		if client.IsNotFound(err) {
			return nil
		}
		return fmt.Errorf("operation returned error: %v\n", err)
	}
	return nil
}

func (client *GCPClient) defaultWaitForZoneOp(
	ctx context.Context, op *compute.Operation, pollInterval time.Duration,
) error {
	err := client.waitForZoneOp(ctx, client.zone, op, pollInterval)
	if err != nil {
		log.Errorf("Operation error: %v\n", err)
		if client.IsNotFound(err) {
			return nil
		}
		return fmt.Errorf("operation returned error: %v\n", err)
	}
	return nil
}

func getRegionForZone(zone string) (string, error) {
	idx := strings.LastIndex(zone, "-")
	if 0 > idx {
		log.Errorf("Invalid zone format: %s\n", zone)
		return "", fmt.Errorf("invalid zone format: %s\n", zone)
	}
	return zone[0:idx], nil
}

func getDefaultComputeService() (*compute.Service, error) {
	gcpClient, err := google.DefaultClient(context.Background(), compute.ComputeScope)
	if nil != err {
		log.Errorf("Couldn't create client: %v\n", err)
		return nil, fmt.Errorf("failed to create client: %v\n", err)
	}

	compute, err := compute.New(gcpClient)
	if nil != err {
		log.Errorf("Couldn't create GCP compute service: %v\n", err)
		return nil, fmt.Errorf("failed to create GCP compute service: %v\n", err)
	}
	return compute, nil
}

func getNewComputeService(credentialsPath string) (*compute.Service, error) {
	if credentialsPath == "" {
		return getDefaultComputeService()
	}
	compute, err := compute.NewService(context.Background(),
		option.WithCredentialsFile(credentialsPath))
	if err != nil {
		log.WithFields(log.Fields{
			"error":           err,
			"credentialsFile": credentialsPath,
		}).Error("Unable to create GCP compute service.")
		return nil, fmt.Errorf("failed to create GCP compute service: %v\n", err)
	}
	return compute, nil
}

func NewGCPClientWithCredentials(
	credentialsPath string, sharedVIP bool,
) (*GCPClient, error) {
	var (
		client GCPClient
		err    error
	)

	client.sharedVIP = sharedVIP

	client.project, err = metadata.ProjectID()
	if nil != err {
		log.Errorf("Couldn't get GCP project: %v\n", err)
		return nil, fmt.Errorf("failed to get GCP project: %v\n", err)
	}

	client.zone, err = metadata.Zone()
	if nil != err {
		log.Errorf("Couldn't get GCP zone: %v\n", err)
		return nil, fmt.Errorf("failed to get GCP zone: %v\n", err)
	}

	client.region, err = getRegionForZone(client.zone)
	if err != nil {
		return nil, err
	}

	clusterInfo, err := client.GetClusterInfo()
	if err != nil {
		return nil, err
	}
	client.cluster = clusterInfo.Name

	client.compute, err = getNewComputeService(credentialsPath)
	if err != nil {
		return nil, err
	}

	client.container, err = getNewContainerServiceWithCredentials(credentialsPath)
	if err != nil {
		return nil, err
	}

	log.WithFields(log.Fields{
		"project": client.project,
		"zone":    client.zone,
		"region":  client.region,
	}).Debug("Initialized GCP client.")

	return &client, nil
}

func NewGCPClient(sharedVIP bool) (*GCPClient, error) {
	return NewGCPClientWithCredentials("", sharedVIP)
}

func NewGCPClientLocal(
	credentialsPath, project, zone, clusterName string, sharedVIP bool,
) (*GCPClient, error) {
	var (
		client GCPClient
		err    error
	)

	client.sharedVIP = sharedVIP
	client.project = project

	client.region, err = getRegionForZone(zone)
	if err != nil {
		return nil, err
	}

	client.zone = zone

	client.cluster = clusterName

	client.compute, err = getNewComputeService(credentialsPath)
	if err != nil {
		return nil, err
	}

	client.container, err = getNewContainerServiceWithCredentials(credentialsPath)
	if err != nil {
		return nil, err
	}

	return &client, nil
}

func NewFakeGCPClient() *GCPClient {
	return &GCPClient{
		sharedVIP: false,
		project:   "quark-dev",
		region:    "us-east1",
		zone:      "us-east1-b",
		compute:   nil, // TODO: see if GCP offers a fake compute service?
	}
}

func (client *GCPClient) CreateLoadBalancerAddress(ctx context.Context, name string, subnet string) error {
	var (
		addr    *compute.Address
		purpose string
	)
	// subnet implies ILB, empty subnet ELB
	if subnet != "" {
		if client.sharedVIP {
			purpose = "SHARED_LOADBALANCER_VIP"
		} else {
			purpose = "GCE_ENDPOINT"
		}
		addr = &compute.Address{
			AddressType: "INTERNAL",
			Subnetwork:  subnet,
			Name:        name,
			Purpose:     purpose,
		}
	} else {
		addr = &compute.Address{
			AddressType: "EXTERNAL",
			Name:        name,
		}
	}

	op, err := client.compute.Addresses.Insert(client.project, client.region, addr).Context(ctx).Do()
	if nil != err {
		if isGCEError(err, "alreadyExists") {
			return nil
		}
		log.WithField("name", name).Errorf("Could not create address: %v\n", err)
		return fmt.Errorf("failed to create address: %v\n", err)
	}

	err = client.waitForRegionalOp(ctx, op)
	if nil != err {
		log.Errorf("Operation error: %v\n", err)
		if isGCEError(err, "alreadyExists") {
			return nil
		}
		return fmt.Errorf("operation returned error: %v\n", err)
	}

	return nil
}

func (client *GCPClient) DeleteLoadBalancerAddress(ctx context.Context, name string) error {
	op, err := client.compute.Addresses.Delete(client.project, client.region, name).Context(ctx).Do()
	if nil != err {
		if client.IsNotFound(err) {
			return nil
		}
		log.WithField("name", name).Errorf("Could not delete address: %v\n", err)
		return fmt.Errorf("failed to delete address: %v\n", err)
	}

	err = client.waitForRegionalOp(ctx, op)
	if nil != err {
		log.Errorf("Operation error: %v\n", err)
		if client.IsNotFound(err) {
			return nil
		}
		return fmt.Errorf("operation returned error: %v\n", err)
	}

	return nil
}

func (client *GCPClient) GetLoadBalancerAddress(ctx context.Context, name string) (string, error) {
	addr, err := client.compute.Addresses.Get(client.project, client.region, name).Context(ctx).Do()
	if nil != err {
		return "", err
	}
	return addr.Address, nil
}

func (client *GCPClient) IsNotFound(err error) bool {
	return isGCEError(err, "notFound")
}

func (client *GCPClient) IsNotReady(err error) bool {
	return isGCEError(err, "resourceNotReady")
}

func (client *GCPClient) ResetInstanceForProviderID(
	ctx context.Context, id string,
) error {
	components := strings.Split(strings.TrimPrefix(id, "gce://"), "/")
	if len(components) != 3 {
		return fmt.Errorf("providerID %s not in gce://<project>/<zone>/<name> "+
			"format", id)
	}
	return client.resetInstanceForProjectZone(ctx, components[2], components[0],
		components[1])
}

func (client *GCPClient) resetInstanceForProjectZone(
	ctx context.Context, name, project, zone string,
) error {
	op, err := compute.NewInstancesService(client.compute).Reset(
		project, zone, name).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project":  project,
			"zone":     zone,
			"instance": name,
			"error":    err}).Error("Unable to issue reset call.")
		return err
	}
	return client.defaultWaitForZoneOp(ctx, op, 5*time.Second)
}
func (client *GCPClient) ResetInstanceForZone(
	ctx context.Context, name, zone string,
) error {
	return client.resetInstanceForProjectZone(ctx, name, client.project, zone)
}

func (client *GCPClient) ResetInstance(ctx context.Context, name string) error {
	return client.ResetInstanceForZone(ctx, name, client.zone)
}

func (client *GCPClient) AddCloudAnnotations(service *corev1.Service, ilbSubnet string, ilbGlobalAccess bool) {
	if service.Annotations == nil {
		service.Annotations = make(map[string]string)
	}
	service.Annotations[quarkconfig.LegacyILBTypeAnnotation] = quarkconfig.ILBInternal
	service.Annotations[quarkconfig.ILBTypeAnnotation] = quarkconfig.ILBInternal

	// Make sure we are not in seclab before adding subnet annotation
	// This is a work around for a google bug with ILB subsetting
	// TODO: could be removed when the google bug is fixed
	if client.project != "netapp-ctl-seclab" {
		service.Annotations[quarkconfig.ILBPUPIAnnotation] = getSubnetNameFromLink(ilbSubnet)
	}

	// Enable Global Access to allow inter-zone access to ILB if requested
	if ilbGlobalAccess {
		service.Annotations[quarkconfig.ILBGlobalAccessAnnotation] = "true"
	} else {
		service.Annotations[quarkconfig.ILBGlobalAccessAnnotation] = "false"
	}
}

func (client *GCPClient) GlobalAccessEnabled(service *corev1.Service) bool {
	return service.Annotations[quarkconfig.ILBGlobalAccessAnnotation] == "true"
}

func (client *GCPClient) GetClusterInfo() (*ClusterInfo, error) {
	var (
		clusterInfo ClusterInfo
		err         error
	)
	clusterInfo.Name, err = metadata.InstanceAttributeValue("cluster-name")
	if err != nil {
		return nil, err
	}
	clusterInfo.Location, err = metadata.InstanceAttributeValue("cluster-location")
	if err != nil {
		return nil, err
	}

	return &clusterInfo, nil
}

func getNewContainerServiceWithCredentials(credentialsPath string) (*container.Service, error) {
	return container.NewService(context.Background(),
		option.WithCredentialsFile(credentialsPath))
}

func (client *GCPClient) waitForContainerOp(
	ctx context.Context, op *container.Operation, pollInterval time.Duration,
) error {
	return wait.Poll(pollInterval, 5*time.Minute, func() (bool, error) {
		opService := container.NewProjectsLocationsOperationsService(client.container)
		opName := fmt.Sprintf("projects/%s/locations/%s/operations/%s",
			client.project, client.region, op.Name)
		getCall := opService.Get(opName)
		getOp, err := getCall.Context(ctx).Do()
		if err != nil {
			log.Errorf("WaitForOp(op: %#v, zone: %#v) failed to poll the operation",
				opName, client.region)
			return false, err
		}
		done, err := containerOpIsDone(getOp)
		return done, err
	})
}

func containerOpIsDone(op *container.Operation) (bool, error) {
	if op == nil || op.Status != "DONE" {
		return false, nil
	}
	if op.StatusMessage != "" {
		return true, fmt.Errorf("operation %v failed: %v", op.Name,
			op.StatusMessage)
	}
	return true, nil
}

func (client *GCPClient) ListZonalNodePools(ctx context.Context) ([]*NodePool, error) {
	zonalNodePoolsService := container.NewProjectsZonesClustersNodePoolsService(client.container)
	nodePoolsListCall := zonalNodePoolsService.List(client.project,
		client.zone, client.cluster)
	response, err := nodePoolsListCall.Context(ctx).Do()
	if err != nil {
		return nil, err
	}

	nodePools := make([]*NodePool, 0)
	for _, pool := range response.NodePools {
		np := &NodePool{
			Name:        pool.Name,
			MachineType: pool.Config.MachineType,
		}
		nodePools = append(nodePools, np)
	}
	return nodePools, nil
}

func (client *GCPClient) ListRegionalNodePools(ctx context.Context) ([]*NodePool, error) {
	nodePoolsService := container.NewProjectsLocationsClustersNodePoolsService(client.container)
	parent := fmt.Sprintf("projects/%s/locations/%s/clusters/%s",
		client.project, client.region, client.cluster)
	nodePoolsListCall := nodePoolsService.List(parent)
	response, err := nodePoolsListCall.Context(ctx).Do()
	if err != nil {
		return nil, err
	}

	nodePools := make([]*NodePool, 0)
	for _, pool := range response.NodePools {
		np := &NodePool{
			Name:        pool.Name,
			MachineType: pool.Config.MachineType,
		}
		nodePools = append(nodePools, np)
	}
	return nodePools, nil
}

func (client *GCPClient) ResizeRegionalNodePool(ctx context.Context,
	nodePool string, newSize int64) error {
	nodePoolsService := container.NewProjectsLocationsClustersNodePoolsService(client.container)
	nodePool = fmt.Sprintf("projects/%s/locations/%s/clusters/%s/nodePools/%s",
		client.project, client.region, client.cluster, nodePool)
	request := container.SetNodePoolSizeRequest{
		Name:      nodePool,
		NodeCount: newSize,
	}

	setPoolSizeCall := nodePoolsService.SetSize(nodePool, &request)
	op, err := setPoolSizeCall.Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project":  client.project,
			"zone":     client.zone,
			"nodePool": nodePool,
			"size":     newSize,
			"error":    err},
		).Error("Unable to set node pool size.")
		return err
	}
	err = client.waitForContainerOp(ctx, op, 1*time.Second)
	if err != nil {
		log.Errorf("Operation error: %v\n", err)
		if client.IsNotFound(err) {
			return nil
		}
		return fmt.Errorf("operation returned error: %v\n", err)
	}
	return nil
}

// FROM 'compute "google.golang.org/api/compute/v0.beta"'
// Insert: Creates a managed instance group using the information that
// you specify in the request. After the group is created, instances in
// the group are created using the specified instance template.
func (client *GCPClient) CreateManagedInstanceGroup(ctx context.Context, group InstanceGroup) error {
	managedInstanceGroup := compute.InstanceGroupManager{
		Name:             group.GetName(),
		BaseInstanceName: group.GetBaseInstanceName(),
		TargetSize:       group.GetSize(),
		Zone:             group.GetZone(),
		InstanceTemplate: group.GetInstanceTemplate(),
	}
	op, err := compute.NewInstanceGroupManagersService(client.compute).Insert(client.project, client.zone, &managedInstanceGroup).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project": client.project,
			"zone":    client.zone,
			"migName": managedInstanceGroup.Name,
			"error":   err}).Error("Unable to issue delete call.")
		return err
	}
	return client.defaultWaitForZoneOp(ctx, op, 1*time.Second)
}

func (client *GCPClient) ListManagedInstanceGroupsForRegionalCluster(
	ctx context.Context,
) ([]InstanceGroupClusterInfo, error) {
	clusterUrl := fmt.Sprintf("projects/%s/locations/%s/clusters/%s",
		client.project, client.region, client.cluster)
	cluster, err := container.NewProjectsLocationsClustersService(
		client.container).Get(clusterUrl).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project":    client.project,
			"region":     client.region,
			"cluster":    client.cluster,
			"clusterURL": clusterUrl,
			"error":      err,
		}).Error("Failed to get cluster")
		return nil, err
	}
	ret := make([]InstanceGroupClusterInfo, 0)
	for _, n := range cluster.NodePools {
		if n.Config == nil {
			log.WithFields(log.Fields{
				"clusterUrl": clusterUrl,
				"name":       n.Name,
			}).Warn("Got nodepool with empty config; skipping.")
			continue
		}
		for _, igu := range n.InstanceGroupUrls {
			// URLS are of the form https://www.googleapis.com/compute/v1/projects/quark-dev-234914/zones/us-east1-b/instanceGroupManagers/gke-cdragga-nvc-clus-n1-standard-32-p-b1af01be-grp
			// The name is the last component of the url, and the zone
			// immediately follows the "zones" components.
			tokens := strings.Split(strings.TrimPrefix(igu, "https://"), "/")
			name := tokens[len(tokens)-1]
			zone := ""
			for i, s := range tokens {
				if s == "zones" && i < len(tokens)-1 {
					zone = tokens[i+1]
					break
				}
			}
			if zone == "" {
				log.WithFields(log.Fields{
					"name":        name,
					"machineType": n.Config.MachineType,
				}).Warn("Got instance group without zone")
			} else {
				ret = append(ret, &gcpInstanceGroupClusterInfo{
					name:        name,
					zone:        zone,
					machineType: n.Config.MachineType,
					nodePool:    n.Name,
				})
			}
		}
	}
	return ret, nil

}

func (client *GCPClient) GetManagedInstanceGroup(managedInstanceGroupName string) (InstanceGroup, error) {
	mig, err := compute.NewInstanceGroupManagersService(client.compute).Get(client.project, client.zone, managedInstanceGroupName).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project": client.project,
			"zone":    client.zone,
			"migName": managedInstanceGroupName,
			"error":   err}).Error("Unable to issue get call.")
		return nil, err
	}
	return &gcpInstanceGroup{mig}, nil
}

func (client *GCPClient) DeleteManagedInstanceGroup(ctx context.Context, managedInstanceGroupName string) error {
	op, err := compute.NewInstanceGroupManagersService(client.compute).Delete(client.project, client.zone, managedInstanceGroupName).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project": client.project,
			"zone":    client.zone,
			"migName": managedInstanceGroupName,
			"error":   err}).Error("Unable to issue delete call.")
		return err
	}
	return client.defaultWaitForZoneOp(ctx, op, 1*time.Second)
}

// FROM 'compute "google.golang.org/api/compute/v0.beta"'
// Resize: Resizes the managed instance group. If you increase the size,
// the group creates new instances using the current instance template.
// If you decrease the size, the group deletes instances. The resize
// operation is marked DONE when the resize actions are scheduled even
// if the group has not yet added or deleted any instances. You must
// separately verify the status of the creating or deleting actions with
// the listmanagedinstances method.
//
func (client *GCPClient) CreateInstanceForGroupInZone(
	ctx context.Context,
	managedGroupName, zone string,
	newSize int64,
) error {
	op, err := compute.NewInstanceGroupManagersService(client.compute).Resize(
		client.project, zone, managedGroupName, newSize,
	).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project": client.project,
			"zone":    zone,
			"migName": managedGroupName,
			"newSize": newSize,
			"error":   err}).Error("Unable to issue resize call.")
		return err
	}
	return client.waitForZoneOp(ctx, zone, op, 1*time.Second)
}

func (client *GCPClient) CreateInstanceForGroup(
	ctx context.Context, managedGroupName string, newSize int64,
) error {
	return client.CreateInstanceForGroupInZone(ctx, managedGroupName,
		client.zone, newSize)
}

// RecreateInstances: Flags the specified instances in the managed
// instance group to be immediately recreated. The instances are deleted
// and recreated using the current instance template for the managed
// instance group. This operation is marked as DONE when the flag is set
// even if the instances have not yet been recreated.
func (client *GCPClient) RecreateInstancesForGroup(ctx context.Context, groupName, zone string, instanceNames ...string) error {
	// The full URL or a partial URL, such as zones/[ZONE]/instances/[INSTANCE_NAME].
	instanceURLs := make([]string, len(instanceNames))
	for i, name := range instanceNames {
		instanceURLs[i] = fmt.Sprintf("zones/%s/instances/%s", zone, name)
	}
	request := &compute.InstanceGroupManagersRecreateInstancesRequest{
		Instances: instanceURLs,
	}
	op, err := compute.NewInstanceGroupManagersService(client.compute).RecreateInstances(client.project, client.zone, groupName, request).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project":   client.project,
			"zone":      client.zone,
			"migName":   groupName,
			"instances": instanceNames,
			"error":     err}).Error("Unable to issue recreate call.")
		return err
	}
	return client.defaultWaitForZoneOp(ctx, op, 1*time.Second)
}

func (client *GCPClient) ListManagedInstancesForGroup(groupName string) ([]ManagedInstance, error) {
	managedInstances, err := compute.NewInstanceGroupManagersService(client.compute).ListManagedInstances(client.project, client.zone, groupName).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project": client.project,
			"zone":    client.zone,
			"migName": groupName,
			"error":   err}).Error("Unable to issue list call.")
		return nil, err
	}
	instances := make([]ManagedInstance, len(managedInstances.ManagedInstances))
	for i := range managedInstances.ManagedInstances {
		instances[i] = &gcpManagedInstance{managedInstances.ManagedInstances[i]}
	}
	return instances, nil
}

// FROM 'compute "google.golang.org/api/compute/v0.beta"'
// DeleteInstances: Flags the specified instances in the managed
// instance group for immediate deletion. The instances are also removed
// from any target pools of which they were a member. This method
// reduces the targetSize of the managed instance group by the number of
// instances that you delete.
func (client *GCPClient) DeleteInstanceFromGroup(ctx context.Context, groupName string, instanceNames ...string) error {
	request := &compute.InstanceGroupManagersDeleteInstancesRequest{
		Instances: instanceNames,
	}
	op, err := compute.NewInstanceGroupManagersService(client.compute).DeleteInstances(client.project, client.zone, groupName, request).Context(ctx).Do()
	if err != nil {
		log.WithFields(log.Fields{
			"project":   client.project,
			"zone":      client.zone,
			"migName":   groupName,
			"instances": instanceNames,
			"error":     err}).Error("Unable to issue delete call.")
		return err
	}
	return client.defaultWaitForZoneOp(ctx, op, 1*time.Second)
}
