/*
 *  Copyright (c) 2021 NetApp
 *  All rights reserved
 */

package regional

import (
	"context"
	"fmt"
	"strings"
	"sync"

	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	storagev1 "k8s.io/api/storage/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/kubernetes"

	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

type RegionalMetadata interface {
	ProcessNewNetAppObj(kubeClientset kubernetes.Interface, volumeName string,
		regionalServiceLevel bool, dataPvcs []*corev1.PersistentVolumeClaim,
		availableZones []string) (string, error)
	DeleteNetAppObj(volumeName string)
	GetFailureDomainForVolume(
		kubeClientset kubernetes.Interface,
		volumeName string,
		dataPvcs []*corev1.PersistentVolumeClaim,
	) ([]string, error)
}

func NewRegionalMetadata(kubeClientset kubernetes.Interface,
	quarkClientset quarkclientset.Interface,
	namespace, cloudProvider string) (RegionalMetadata, error) {
	switch cloudProvider {
	case quarkconfig.CloudGCP:
		return newGCPRegionalMetadata(kubeClientset, quarkClientset, namespace)
	case quarkconfig.CloudAzure:
		return newAzureRegionalMetadata()
	default:
		return nil, fmt.Errorf("unsupported cloud provider: %s", cloudProvider)
	}
}

// GKE definitions
type gcpRegionalMetadata struct {
	// Map of failure domains (zone/zone combinations) to count of volumes using
	// it
	zonalFailureDomainUsage    map[string]int
	regionalFailureDomainUsage map[string]int
	failureDomainForVolume     map[string]string
	failureDomainUsageLock     sync.Mutex
}

func newGCPRegionalMetadata(
	kubeClientset kubernetes.Interface,
	quarkClientset quarkclientset.Interface,
	namespace string,
) (*gcpRegionalMetadata, error) {
	ret := &gcpRegionalMetadata{
		zonalFailureDomainUsage:    map[string]int{},
		regionalFailureDomainUsage: map[string]int{},
		failureDomainForVolume:     map[string]string{},
	}

	// Not grabbing the mutex here, because we assume that the watchers haven't
	// started yet.
	storageClasses, err := kubeClientset.StorageV1().StorageClasses().List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("unable to list storage classes: %v", err)
	}
	regionalFailureDomains := []string{}
	zonalFailureDomains := []string{}
	for _, sc := range storageClasses.Items {
		// We don't care about storage classes that don't specify allowed zones.
		if len(sc.AllowedTopologies) == 0 {
			continue
		}
		failureDomain, regional := getFailureDomainForStorageClass(&sc)
		if failureDomain == "" {
			log.WithFields(log.Fields{
				"scName": sc.Name,
			}).Warn("Got storage class with allowed topologies that is not based " +
				"on zone.")
			continue
		}
		if regional {
			if _, ok := ret.regionalFailureDomainUsage[failureDomain]; !ok {
				regionalFailureDomains = append(regionalFailureDomains, failureDomain)
			}
			ret.regionalFailureDomainUsage[failureDomain] = 0
		} else {
			if _, ok := ret.zonalFailureDomainUsage[failureDomain]; !ok {
				zonalFailureDomains = append(zonalFailureDomains, failureDomain)
			}
			ret.zonalFailureDomainUsage[failureDomain] = 0
		}
	}
	log.WithFields(log.Fields{
		"regionalFailureDomains": strings.Join(regionalFailureDomains, ", "),
		"zonalFailureDomains":    strings.Join(zonalFailureDomains, ", "),
	}).Debug("Got failure domains for cluster.")
	return ret, nil
}

// Returns failure domain for a storage pool or volume volume and error,
// if appropriate.
// Assigns the volume a new one if it doesn't already existing
func (r *gcpRegionalMetadata) ProcessNewNetAppObj(
	kubeClientset kubernetes.Interface,
	name string,
	regionalServiceLevel bool,
	dataPvcs []*corev1.PersistentVolumeClaim,
	availableZones []string,
) (string, error) {
	var (
		failureDomain string
		regional      bool
		err           error
	)

	r.failureDomainUsageLock.Lock()
	defer r.failureDomainUsageLock.Unlock()

	if failureDomain, ok := r.failureDomainForVolume[name]; ok {
		return failureDomain, nil
	}
	if len(dataPvcs) == 0 {
		regional = regionalServiceLevel
		usageMap := r.zonalFailureDomainUsage
		if regionalServiceLevel {
			usageMap = r.regionalFailureDomainUsage
		}
		if len(availableZones) == quarkconfig.AvailableZonesGCP {
			// Find the failure domain that matches the specified zones
			for fd, _ := range usageMap {
				if fd == util.GetFailureDomainString(availableZones) {
					failureDomain = fd
					break
				}
			}
			if failureDomain == "" {
				return failureDomain, fmt.Errorf("invalid failure domain '%s'", util.GetFailureDomainString(availableZones))
			}
		} else {
			// We don't have a failure domain already, so get the one with the fewest
			// volumes.
			minVolumes := 10000 // TODO: Create a constant for the max volume count.
			for fd, count := range usageMap {
				if count < minVolumes {
					minVolumes = count
					failureDomain = fd
				}
			}
		}
	} else {
		failureDomain, regional, err = getFailureDomainForPVC(kubeClientset,
			dataPvcs[0], name)
		if err != nil {
			return "", err // the function takes care of the error message
		}
		if regional != regionalServiceLevel {
			log.WithFields(log.Fields{
				"regional":             regional,
				"regionalServiceLevel": regionalServiceLevel,
				"failureDomain":        failureDomain,
			}).Warn("Mismatch between storage class failure domain type and " +
				"service level.")
			// note that because we check regional when updating the usageMap
			// below, we're deferring to the storageclass here.
			// TODO: Error on this?
		}
	}
	usageMap := r.zonalFailureDomainUsage
	if regional {
		usageMap = r.regionalFailureDomainUsage
	}
	usageMap[failureDomain]++
	r.failureDomainForVolume[name] = failureDomain
	log.WithFields(log.Fields{
		"volume":        name,
		"regional":      regional,
		"failureDomain": failureDomain,
		"zonalUsage":    r.zonalFailureDomainUsage,
		"regionalUsage": r.regionalFailureDomainUsage,
	}).Info("Updated failure domain usage for new volume.")
	return failureDomain, nil
}

func (r *gcpRegionalMetadata) DeleteNetAppObj(name string) {
	var regional bool

	r.failureDomainUsageLock.Lock()
	defer r.failureDomainUsageLock.Unlock()

	failureDomain, ok := r.failureDomainForVolume[name]
	if !ok {
		return
	}
	delete(r.failureDomainForVolume, name)
	if _, ok = r.zonalFailureDomainUsage[failureDomain]; ok {
		r.zonalFailureDomainUsage[failureDomain]--
		regional = false
	} else if _, ok = r.regionalFailureDomainUsage[failureDomain]; ok {
		r.regionalFailureDomainUsage[failureDomain]--
		regional = true
	} else {
		log.WithFields(log.Fields{
			"volume":        name,
			"failureDomain": failureDomain,
		}).Warn("Unable to find volume failure domain in either the zonal or " +
			"regional usage maps when deleting. Usage counts may be off.")
	}
	log.WithFields(log.Fields{
		"volume":        name,
		"regional":      regional,
		"failureDomain": failureDomain,
		"zonalUsage":    r.zonalFailureDomainUsage,
		"regionalUsage": r.regionalFailureDomainUsage,
	}).Info("Updated failure domain usage for deleted volume.")
}

func (r *gcpRegionalMetadata) GetFailureDomainForVolume(
	kubeClientset kubernetes.Interface,
	volumeName string,
	dataPvcs []*corev1.PersistentVolumeClaim,
) ([]string, error) {
	r.failureDomainUsageLock.Lock()
	defer r.failureDomainUsageLock.Unlock()

	failureDomain, ok := r.failureDomainForVolume[volumeName]
	if ok {
		return util.GetFailureDomainZones(failureDomain), nil
	}
	failureDomain, _, err := getFailureDomainForPVC(kubeClientset,
		dataPvcs[0], volumeName)
	if err != nil {
		return []string{}, err // the function takes care of the error message
	}
	return util.GetFailureDomainZones(failureDomain), nil
}

// Azure definitions
// Volumes provisioned with Azure ZRS (Zone Redundant Storage) are available across all three zones.
// These volumes can be mounted (in R/W mode) from a pod/container in any of the zone
// This gives us an advantage over GKE clusters where the regional PDs are available only in zone pairs (two zones).
// As and when zone selection/ zone priority becomes a hard requirement for Azure the logic for regionalMetadata can be filled in.
type azureRegionalMetadata struct {
}

func newAzureRegionalMetadata() (*azureRegionalMetadata, error) {
	return nil, nil
}

func (r *azureRegionalMetadata) ProcessNewNetAppObj(
	kubeClientset kubernetes.Interface,
	name string,
	regionalServiceLevel bool,
	dataPvcs []*corev1.PersistentVolumeClaim,
	availableZones []string,
) (string, error) {
	return "", nil
}
func (r *azureRegionalMetadata) DeleteNetAppObj(name string) {

}

func (r *azureRegionalMetadata) GetFailureDomainForVolume(
	kubeClientset kubernetes.Interface,
	volumeName string,
	dataPvcs []*corev1.PersistentVolumeClaim,
) ([]string, error) {
	return []string{}, nil
}

// Failure domains helper functions

// Returns:
// Failure domain string if failure domain is present; if not, returns ""
// Regional: true if regional, false if zonal
func getFailureDomainForStorageClass(sc *storagev1.StorageClass) (string, bool) {
	for _, topologySelector := range sc.AllowedTopologies {
		for _, req := range topologySelector.MatchLabelExpressions {
			if req.Key == quarkconfig.ZoneTopologyKey {
				regional := len(req.Values) > 1
				failureDomain := util.GetFailureDomainString(req.Values)
				return failureDomain, regional
			}
		}
	}
	return "", false
}

// This should only be called on PVCs in regional clusters.
func getFailureDomainForPVC(
	kubeClientset kubernetes.Interface,
	pvc *corev1.PersistentVolumeClaim,
	volumeName string,
) (string, bool, error) {
	if pvc.Spec.StorageClassName == nil {
		return "", false, fmt.Errorf(
			"got nil storage class for data pvc %s for volume "+
				"%s", pvc.Name, volumeName)
	}
	sc, err := kubeClientset.StorageV1().StorageClasses().Get(
		context.TODO(), *pvc.Spec.StorageClassName, metav1.GetOptions{})
	if err != nil {
		return "", false,
			fmt.Errorf("unable to get storage class for existing volume: "+
				"%v", err)
	}
	if len(sc.AllowedTopologies) == 0 {
		return "", false,
			fmt.Errorf("storage class %s for volume %s has no "+
				"AllowedTopologies defined in regional cluster", sc.Name, volumeName)
	}
	failureDomain, regional := getFailureDomainForStorageClass(sc)
	if failureDomain == "" {
		return "", false,
			fmt.Errorf("storage class %s for volume %s did not have a "+
				"failure domain in a regional cluster", sc.Name, volumeName)
	}
	return failureDomain, regional, nil
}
