package v2

import (
	"context"
	"fmt"
	"strings"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"

	_ "k8s.io/client-go/plugin/pkg/client/auth"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"

	"quark.netapp.io/volume-controller/pkg/nvc"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

var (
	// supporting smaller storagePool for upgrade and internal configs
	minStoragePoolSize      resource.Quantity = resource.MustParse("1Gi")
	maxStoragePoolSize      resource.Quantity = resource.MustParse("200Ti")
	maxStoragePoolSizeAzure resource.Quantity = resource.MustParse("100Ti")
)

const (
	versionSPUpgradeErr = "Failed to upgrade NetAppStoragePool"
	upgradeSPAbortErr   = "Failed to abort NetAppStoragePoolUpgrade"
)

type MigrateZoneAbortZoneFailureError struct {
}

func (e MigrateZoneAbortZoneFailureError) Error() string {
	return "Unable to abort migration caused by zone failure."
}

type MigrateZoneAbortNoMigrationFailureError struct {
}

func (e MigrateZoneAbortNoMigrationFailureError) Error() string {
	return "No zone migration to abort."
}

type GlobalAccessUpdateAbortNoUpdateError struct{}

func (e GlobalAccessUpdateAbortNoUpdateError) Error() string {
	return "global access update not in progress"
}

type GlobalAccessAbortInProgressError struct{}

func (e GlobalAccessAbortInProgressError) Error() string {
	return "global access abort in progress"
}

// NetAppStoragePoolOnline returns "true" when the Quark pod is ready to serve data.
func (client *NVCClient) NetAppStoragePoolOnline(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	for _, cond := range storagePool.Status.Conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppStoragePoolSMBConfigured returns "true" if SMB server id configured in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBConfigured(storagePool)
}

// Get the storage pool status
func (client *NVCClient) NetAppStoragePoolStatus(storagePoolUUID string) (bool, string) {
	storagePoolLifeCycleState := false
	storagePoolLifeCycleStateDetails := "Storagepool is in error state"
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		storagePoolLifeCycleStateDetails = "Storagepool doesn't exists"
	} else if err != nil {
		storagePoolLifeCycleStateDetails = "Storagepool get failed"
	} else if client.NetAppStoragePoolOnline(storagePool) {
		storagePoolLifeCycleState = true
		storagePoolLifeCycleStateDetails = "Storagepool is available"
	}
	return storagePoolLifeCycleState, storagePoolLifeCycleStateDetails
}

// Initiate NetAppStoragePool version upgrade
func (client *NVCClient) NetAppStoragePoolUpgrade(storagePoolUUID, version string) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetappStoragePool %s was not found for upgrade", storagePoolUUID)
	}
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("%s: %s", versionSPUpgradeErr, err)
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("%s: no quarks found", versionSPUpgradeErr)
	}
	quark := quarks[0]
	if quark.Status.Version != version {
		return nil, fmt.Errorf("%s: quark is not updated to version %s", versionSPUpgradeErr, version)
	}
	// initiate upgrade when Quark is Ready and updated with required version for storagePool upgrade
	storagePool.Spec.QuarkVersion = version
	newStoragePool, err := client.netAppStoragePoolUpdate(storagePool)

	return newStoragePool, err
}

// Abort NetAppStoragePool version upgrade
func (client *NVCClient) NetAppStoragePoolUpgradeAbort(storagePoolUUID string) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetappStoragePool %s was not found for aborting upgrade", storagePoolUUID)
	}
	if storagePool.Spec.QuarkVersion == "" {
		return nil, fmt.Errorf("%s: NetAppStoragePool Spec version is missing", upgradeSPAbortErr)
	}
	// check if abort is possible
	if storagePool.Status.QuarkVersion != "" && storagePool.Spec.QuarkVersion != storagePool.Status.QuarkVersion {
		err := client.checkUpgradeAbortStoragePool(storagePool.Status.Conditions)
		if err != nil {
			return nil, fmt.Errorf("%s: %s", upgradeSPAbortErr, err.Error())
		}
	}
	// revert to old version, in order to signal abort
	storagePool.Spec.QuarkVersion = storagePool.Status.QuarkVersion
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		_, staleError := err.(StaleVersionError)
		if staleError {
			return nil, err
		}
		return nil, fmt.Errorf("%s: %s", upgradeSPAbortErr, err.Error())
	}
	return storagePool, nil
}

// returns error if abort cannot proceed
func (client *NVCClient) checkUpgradeAbortStoragePool(conditions []quarkv1alpha1.NetAppStoragePoolCondition) error {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltSucceeded {
			// do not proceed with abort if halt succeeded
			return fmt.Errorf("deployment is being updated for storagePool upgrade")
		}
	}
	return nil
}

func (client *NVCClient) NetAppStoragePoolGet(storagePoolUUID string) (*quarkv1alpha1.NetAppStoragePool, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.TODO(), storagePoolUUID, metav1.GetOptions{})
}

func (client *NVCClient) NetAppStoragePoolGetVersion(storagePoolUUID string) (string, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return "", fmt.Errorf("failed to get the NetappStoragePool version: %s", err.Error())
	}
	return storagePool.Status.QuarkVersion, nil
}

// returns conditions from NetAppStoragePool CR status
func (client *NVCClient) NetAppStoragePoolGetConditions(storagePoolUUID string) ([]quarkv1alpha1.NetAppStoragePoolCondition, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetappStoragePool: %s", err.Error())
	}
	return storagePool.Status.Conditions, nil
}

func (client *NVCClient) NetAppStoragePoolGetAll() ([]*quarkv1alpha1.NetAppStoragePool, error) {
	storagePools := make([]*quarkv1alpha1.NetAppStoragePool, 0)
	storagePoolsList, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	for i := range storagePoolsList.Items {
		storagePools = append(storagePools, &storagePoolsList.Items[i])
	}
	return storagePools, err
}

func (client *NVCClient) NetAppStoragePoolGetAvailableSpace(storagePoolUUID string) (availableSpace int64, err error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil || storagePool == nil {
		return availableSpace, fmt.Errorf("failed to get the NetappStoragePool: %s", err.Error())
	}
	if !client.NetAppStoragePoolOnline(storagePool) {
		return availableSpace, fmt.Errorf("NetappStoragePool %s is not online yet", storagePoolUUID)
	}
	volumes, err := client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return availableSpace, err
	}
	var usedStoragePoolSize int64 = 0
	for _, volume := range volumes.Items {
		usedStoragePoolSize = usedStoragePoolSize + volume.Status.Size.Value()
	}
	availableSpace = storagePool.Status.Size.Value() - usedStoragePoolSize
	return availableSpace, nil
}

func (client *NVCClient) NetAppStoragePoolGetSMBStatus(storagePoolUUID string) (configured bool, err error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil || storagePool == nil {
		return false, fmt.Errorf("failed to get the NetappStoragePool: %s", err.Error())
	}
	if !client.NetAppStoragePoolOnline(storagePool) {
		return false, fmt.Errorf("NetappStoragePool %s is not online yet", storagePoolUUID)
	}
	return client.NetAppStoragePoolSMBConfigured(storagePool), nil
}

func (client *NVCClient) NetAppStoragePoolGetAllSMB() ([]*quarkv1alpha1.NetAppStoragePool, error) {
	storagePools := make([]*quarkv1alpha1.NetAppStoragePool, 0)
	storagePoolsList, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	for i := range storagePoolsList.Items {
		if storagePoolsList.Items[i].Spec.SMBServerConfiguration.Netbios != "" {
			storagePools = append(storagePools, &storagePoolsList.Items[i])
		}
	}
	return storagePools, err
}

func (client *NVCClient) NetAppVolumeListGetByStoragePool(storagePoolUUID string) (*quarkv1alpha1.NetAppVolumeList, error) {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.StoragePoolLabel] = storagePoolUUID
	labelSelector := fmt.Sprintf("quark.netapp.io/storagePool=%s", storagePoolUUID)
	listOptions := metav1.ListOptions{
		LabelSelector: labelSelector,
	}
	volumesList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).List(
		context.TODO(), listOptions)
	if err != nil {
		return nil, err
	}
	return volumesList, err
}

func (client *NVCClient) NetAppStoragePoolDelete(storagePoolUUID string, user, password string) error {
	volumes, err := client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return err
	}

	if len(volumes.Items) > 0 {
		return fmt.Errorf("failed to delete Storage pool: still in use")
	}

	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return fmt.Errorf("failed to get the NetappStoragePool: %s", err.Error())
	}
	if storagePool.Spec.SMBServerConfiguration.Netbios != "" {
		storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
		if err != nil {
			return fmt.Errorf("failed to get the NetappStoragePool: %s", err.Error())
		}
		_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, util.GetVolDeleteADSecretName(storagePool.Name), user, password)
		if err != nil {
			return err
		}
	}
	return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Delete(context.TODO(), storagePoolUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) netAppStoragePoolUpdate(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Update(context.TODO(), storagePool, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return storagePool, nil
}

// NetAppStoragePoolResizeAbort aborts the current StoragePool-Resize operation
func (client *NVCClient) NetAppStoragePoolResizeAbort(storagePoolUUID string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.TODO(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	// Reset the new size in spec. The storagePool error conditions can then be cleared by Podrick.
	storagePool.Spec.Size = storagePool.Status.Size
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	return storagePool, err
}

// Note that if the migration has completed by the time this operation is
// issued, the operation will do nothing
func (client *NVCClient) NetAppStoragePoolMigrateZoneAbort(
	storagePoolUUID string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(
		client.nvcNamespace).Get(
		context.TODO(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	if storagePool.Spec.Zone == storagePool.Status.Zone {
		return nil, MigrateZoneAbortNoMigrationFailureError{}
	}

	if storagePool.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseFailure {
		// return an error here, since we can't abort zone migrations due to failure
		return nil, MigrateZoneAbortZoneFailureError{}
	}
	storagePool.Spec.Zone = storagePool.Status.Zone
	storagePool.Spec.LastZoneMigrationCause = storagePool.Status.LastZoneMigrationCause
	storagePool.Spec.ZoneMigrationCount = storagePool.Status.ZoneMigrationCount
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	return storagePool, err
}

// NetAppStoragePoolResized returns "true" when the data PVCs are correctly resized.
func (client *NVCClient) NetAppStoragePoolResized(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return sp.Status.Size.Cmp(sp.Spec.Size) == 0
}

func (client *NVCClient) NetAppStoragePoolMigratedZone(sp *quarkv1alpha1.NetAppStoragePool) bool {
	// Only acknowledge a zone migration if it wasn't initiated as part of a zone
	// failure.
	return sp.Status.Zone == sp.Spec.Zone && sp.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseUser
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdated(
	sp *quarkv1alpha1.NetAppStoragePool,
) bool {
	// If global access enablement is being aborted, the spec and status fields
	// will be equal, but GlobalAccessAborting will be true. We only want to
	// declare this to be finalized once GlobalAccessAborting is set to false.
	return sp.Status.GlobalAccess == sp.Spec.GlobalAccess &&
		!sp.Spec.GlobalAccessAborting
}

// NetAppStoragePoolADDNSUpdated returns "true" when AD DNSIPs are applied to the storagePool.
func (client *NVCClient) NetAppStoragePoolADDNSUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsEqualStrings(sp.Spec.ADDNSIPs, sp.Status.ADDNSIPs)
}

// NetAppStoragePoolADPasswordResetDone returns "true" when machine-account password reset is successful.
func (client *NVCClient) NetAppStoragePoolADPasswordResetDone(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return (sp.Status.SMBPasswordResetCount == sp.Spec.SMBPasswordResetCount)
}

// NetAppStoragePoolSMBADConfigurationUpdated returns "true" when AD configurations updated for the storagePool.
func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return (sp.Status.SMBADConfiguration == sp.Spec.SMBADConfiguration)
}

// NetAppStoragePoolServerConfigurationUpdated returns "true" when SMB server configurations are updated for the storagePool.
func (client *NVCClient) NetAppStoragePoolSMBServerConfigurationUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsSMBServerConfigurationEqual(sp)
}

// NetAppStoragePoolUpgraded returns "true" when StoragePool version upgrade is successful.
func (client *NVCClient) NetAppStoragePoolUpgraded(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	conditions := storagePool.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltSucceeded || cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltRequested || cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeFailed {
			// return false if another upgrade is in progress
			return false
		}
	}
	return (storagePool.Status.QuarkVersion == storagePool.Spec.QuarkVersion)

}

// NetAppStoragePoolDelete is used to signal the deletion of a NetAppStoragePool.
type NetAppStoragePoolDelete struct {
	StoragePool *quarkv1alpha1.NetAppStoragePool
}

// NetAppStoragePoolUpdate is used to signal the most recent state of a NetAppStoragePool and the
// most recent changes.
type NetAppStoragePoolUpdate struct {
	StoragePool            *quarkv1alpha1.NetAppStoragePool
	Status                 *NetAppStoragePoolUpdateStatus
	UpdatedConditions      []quarkv1alpha1.NetAppStoragePoolCondition
	LastZoneMigrationCause string
	Error                  error
}

// NetAppStoragePoolUpdateStatus captures the status of any pending updates.
type NetAppStoragePoolUpdateStatus struct {
	SizeUpdated                   bool
	ZoneUpdated                   bool
	GlobalAccessUpdated           bool
	ADDNSUpdated                  bool
	SMBADConfigurationUpdated     bool
	SMBServerConfigurationUpdated bool
	ADPasswordReset               bool
	QuarkVersionUpdated           bool
}

// NetAppStoragePoolUpdated returns "true" when the actual state has converged to the
// desired state. If not, it will return an object describing what actions are pending.
func (client *NVCClient) NetAppStoragePoolUpdated(storagePool *quarkv1alpha1.NetAppStoragePool) (bool, *NetAppStoragePoolUpdateStatus, error) {
	statusConverged := true

	status := &NetAppStoragePoolUpdateStatus{
		SizeUpdated:                   true,
		ZoneUpdated:                   true,
		GlobalAccessUpdated:           true,
		ADDNSUpdated:                  true,
		SMBADConfigurationUpdated:     true,
		SMBServerConfigurationUpdated: true,
		ADPasswordReset:               true,
		QuarkVersionUpdated:           true,
	}
	// Check size
	if !client.NetAppStoragePoolResized(storagePool) {
		status.SizeUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolMigratedZone(storagePool) {
		status.ZoneUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolGlobalAccessUpdated(storagePool) {
		status.GlobalAccessUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolADDNSUpdated(storagePool) {
		status.ADDNSUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBADConfigurationUpdated(storagePool) {
		status.SMBADConfigurationUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBServerConfigurationUpdated(storagePool) {
		status.SMBServerConfigurationUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolADPasswordResetDone(storagePool) {
		status.ADPasswordReset = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolUpgraded(storagePool) {
		status.QuarkVersionUpdated = false
		statusConverged = false
	}
	// Return the status of any potentially pending updates.
	if statusConverged {
		return true, status, nil
	} else {
		return false, status, nil
	}
}

func (client *NVCClient) handleStoragePoolDeletes(sp *quarkv1alpha1.NetAppStoragePool) {
	client.storagePoolDeleteChannel <- &NetAppStoragePoolDelete{sp}
}

func (client *NVCClient) handleStoragePoolUpdates(oldSp *quarkv1alpha1.NetAppStoragePool, newSp *quarkv1alpha1.NetAppStoragePool) {
	sendEvent := false

	// We send events when any of the NetAppStoragePool has converged to the desired state.
	_, newStatus, _ := client.NetAppStoragePoolUpdated(newSp)
	_, oldStatus, _ := client.NetAppStoragePoolUpdated(oldSp)
	if newStatus != oldStatus {
		log.WithFields(log.Fields{
			"storagePool": newSp.Name,
			"newStatus":   newStatus,
		}).Debug("NetApp Storage pool state has been updated.")
		sendEvent = true
	}
	// We send events when the NetStoragePool's conditions have changed.
	if len(newSp.Status.Conditions) != len(oldSp.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"storagePool": newSp.Name,
			"conditions":  newSp.Status.Conditions,
		}).Debug("NetApp Storage pool has updated conditions.")
	} else {
		for i, cond := range newSp.Status.Conditions {
			if oldSp.Status.Conditions[i].Type != cond.Type || oldSp.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"storagePool": newSp.Name,
					"conditions":  newSp.Status.Conditions,
				}).Debug("NetApp Storagepool has updated conditions.")
				break
			}
		}
	}
	if newSp.Spec.LastZoneMigrationCause != oldSp.Spec.LastZoneMigrationCause {
		sendEvent = true
	}

	if sendEvent {
		client.storagePoolUpdateChannel <- &NetAppStoragePoolUpdate{
			StoragePool:            newSp,
			Status:                 newStatus,
			UpdatedConditions:      newSp.Status.Conditions,
			LastZoneMigrationCause: newSp.Spec.LastZoneMigrationCause,
		}
	}
}

func (client *NVCClient) validateStoragePoolSize(size resource.Quantity) (err error) {
	var (
		maxStoragePoolSizeAllowed resource.Quantity
	)
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	if quark.Spec.Cloud == quarkconfig.CloudAzure {
		maxStoragePoolSizeAllowed = maxStoragePoolSizeAzure
	} else {
		maxStoragePoolSizeAllowed = maxStoragePoolSize
	}
	minStoragePoolSizeDifference := size.Value() - minStoragePoolSize.Value()
	maxStoragePoolSizeDifference := size.Value() - maxStoragePoolSizeAllowed.Value()
	if minStoragePoolSizeDifference < 0 || maxStoragePoolSizeDifference > 0 || minStoragePoolSizeDifference%incrementalSize.Value() != 0 {
		return fmt.Errorf("size=%s, storage pool size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minStoragePoolSize.String(), maxStoragePoolSizeAllowed.String(), incrementalSize.String())
	}
	return nil
}

func (client *NVCClient) NetAppStoragePoolSizeUpdate(sp *quarkv1alpha1.NetAppStoragePool, newSizeString string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	size, err := resource.ParseQuantity(newSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}

	err = client.validateStoragePoolSize(size)
	if err != nil {
		return nil, err
	}
	if sp.Spec.Size.Value() > size.Value() {
		return nil, fmt.Errorf("storage pool size shrink operation is not allowed. Current storagePool size is %s, newsize=%s", sp.Spec.Size.String(), size.String())
	}
	sp.Spec.Size = size

	sp, err = client.netAppStoragePoolUpdate(sp)

	return sp, err
}

func (client *NVCClient) NetAppStoragePoolMigrateZone(
	sp *quarkv1alpha1.NetAppStoragePool, zone string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if sp.Spec.Zone == zone {
		// We already did the migration; don't update.
		return sp, nil
	}
	// TODO: Validate zone choice here once zone selection is in
	sp.Spec.Zone = zone
	sp.Spec.LastZoneMigrationCause = quarkconfig.ZoneMigrationCauseUser
	sp.Spec.ZoneMigrationCount = sp.Spec.ZoneMigrationCount + 1
	// This is subject to clock skew, but there's not a good way around this.
	now := metav1.Now()
	sp.Spec.LastZoneMigrationTime = &now
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdate(
	sp *quarkv1alpha1.NetAppStoragePool, enabled bool,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if sp.Spec.GlobalAccessAborting {
		return nil, GlobalAccessAbortInProgressError{}
	}
	sp.Spec.GlobalAccess = enabled
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessSet(
	sp *quarkv1alpha1.NetAppStoragePool, enabled bool,
) *quarkv1alpha1.NetAppStoragePool {
	sp.Spec.GlobalAccess = enabled
	return sp
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdateAbort(
	storagePoolUUID string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	sp, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(
		client.nvcNamespace).Get(
		context.TODO(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	if sp.Status.GlobalAccess == sp.Spec.GlobalAccess {
		return nil, GlobalAccessUpdateAbortNoUpdateError{}
	}
	sp.Spec.GlobalAccess = sp.Status.GlobalAccess
	if nvc.HasFinalizer(sp.Finalizers,
		quarkconfig.ADClientAddressFinalizer) && sp.Status.ADClientIP == "" {
		// We're in the process of first time enablement; mark that we're
		// aborting
		sp.Spec.GlobalAccessAborting = true
	}
	return client.netAppStoragePoolUpdate(sp)
}

// Reset Machine Account password for SMB storagePool
func (client *NVCClient) NetAppStoragePoolSMBPasswordReset(storagePoolUUID string, user, password string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	secretName := util.GetVolPwdResetADSecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": storagePool.Name,
			"secret": secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	storagePool.Spec.SMBPasswordResetCount = storagePool.Spec.SMBPasswordResetCount + 1
	newStoragePool, err := client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		delErr := client.SecretDelete(secretName)
		if delErr != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
				"error":       delErr,
			}).Error("Failed to delete secret.")
		}

		return nil, err
	}
	return newStoragePool, err
}

// Abort machine-account password reset operation.
// SDE should abort the password reset operation if it times out.
// Abort will delete the secret and avoid retrying operation.
func (client *NVCClient) NetAppStoragePoolADPasswordResetAbort(storagePoolUUID string) (sp *quarkv1alpha1.NetAppStoragePool, err error) {
	secretName := util.GetVolPwdResetADSecretName(storagePoolUUID)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	sp, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	sp.Spec.SMBPasswordResetCount = sp.Status.SMBPasswordResetCount
	sp, err = client.netAppStoragePoolUpdate(sp)

	return sp, err
}

// Update DNS IP for AD storagePool
func (client *NVCClient) NetAppStoragePoolADDNSUpdate(sp *quarkv1alpha1.NetAppStoragePool, dnsIPs []string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.ADDNSIPs = dnsIPs
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return
}

func (client *NVCClient) NetAppStoragePoolADDNSUpdateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.ADDNSIPs = storagePool.Status.ADDNSIPs
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return
}

// Update EncryptedCommunication for AD storagePool
func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdate(sp *quarkv1alpha1.NetAppStoragePool, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return
}

func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBADConfiguration = storagePool.Status.SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return
}

func (client *NVCClient) NetAppStoragePoolSMBConfigCreate(sp *quarkv1alpha1.NetAppStoragePool, user, password string, ADDNSIPs []string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}

	if storagePool.Status.SMBServerConfiguration.Netbios != "" {
		if storagePool.Status.SMBServerConfiguration != SMBServerConfiguration || storagePool.Status.SMBADConfiguration != SMBADConfiguration ||
			!util.IsEqualStrings(storagePool.Status.ADDNSIPs, ADDNSIPs) {
			return nil, fmt.Errorf("updating CIFS Server config is not allowed: Server already congfigured")
		}
	}

	secretName := util.GetVolCreateADSecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecret(secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}
	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool.Spec.ADDNSIPs = ADDNSIPs
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return
}

func (client *NVCClient) NetAppStoragePoolSMBConfigCreateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	secretName := util.GetVolCreateADSecretName(sp.Name)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.ADDNSIPs = storagePool.Status.ADDNSIPs
	storagePool.Spec.SMBServerConfiguration = storagePool.Status.SMBServerConfiguration
	storagePool.Spec.SMBADConfiguration = storagePool.Status.SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return
}

func (client *NVCClient) CheckAvailableZones(serviceLevel ServiceLevel, availableZones []string) (bool, error) {
	var err error

	switch serviceLevel {
	case ServiceLevelStandard:
		if len(availableZones) > 0 {
			return false, fmt.Errorf("invalid value for availableZones; "+
				"at the standard service level expect availableZones to be empty, "+
				"found %d zones specified", len(availableZones))
		}

	case ServiceLevelRegionalStandard:
		if len(availableZones) == 0 {
			// We can use the default behavior; no reason to look up the cloud
			// provider
			return true, nil
		}
		// Need to know what cloud provider
		quarks, err := client.NetAppQuarkGetAll()
		if err != nil {
			return false, fmt.Errorf("unable to get Cloud provider, error retrieving quark CR: %v", err)
		}
		if len(quarks) == 0 {
			return false, fmt.Errorf("unable to get Cloud provider, quark CR not found")
		}
		quark := quarks[0]
		if quark.Spec.Cloud == quarkconfig.CloudGCP {
			if len(availableZones) != quarkconfig.AvailableZonesGCP {
				return false, fmt.Errorf("invalid number of availableZones, "+
					"for Cloud provider, %s, expect %d, received %d",
					quark.Spec.Cloud, quarkconfig.AvailableZonesGCP, len(availableZones))
			}
		} else if quark.Spec.Cloud == quarkconfig.CloudAzure {
			if len(availableZones) != quarkconfig.AvailableZonesAzure {
				return false, fmt.Errorf("invalid number of availableZones, "+
					"for Cloud provider, %s, expect %d, received %d",
					quark.Spec.Cloud, quarkconfig.AvailableZonesAzure, len(availableZones))
			}
		}
	}
	return true, err
}

// ConstructRegionalNetAppStoragePoolWithSMBConfig() does validations on availableZones before calling
// ConstructNetAppStoragePoolWithSMBConfig.  It then sets Zone and AvailableZones fields of storagePool.Spec
func (client *NVCClient) ConstructRegionalNetAppStoragePoolWithSMBConfig(ownerID, displayName, sizeString,
	storagePoolUUID, zone string, availableZones []string, serviceLevel ServiceLevel, ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {

	zonesOk, err := client.CheckAvailableZones(serviceLevel, availableZones)
	if !zonesOk {
		return nil, err
	}

	storagePool, err = client.ConstructNetAppStoragePoolWithSMBConfig(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel, ADDNSIPs, SMBServerConfiguration, SMBADConfiguration)
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}

	// Set Zone and availableZones
	storagePool.Spec.Zone = zone
	storagePool.Spec.AvailableZones = availableZones
	return storagePool, err
}

func (client *NVCClient) ConstructNetAppStoragePoolWithSMBConfig(ownerID, displayName, sizeString, storagePoolUUID string, serviceLevel ServiceLevel, ADDNSIPs []string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel)
	if storagePool == nil || err != nil {
		log.Fatalf("Failed to construct StoragePool")
		return
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool.Spec.ADDNSIPs = ADDNSIPs
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	return storagePool, err
}

// ConstructRegionalNetAppStoragePool() does validations on availableZones before calling
// ConstructNetAppStoragePool.  It then sets Zone and AvailableZones fields of storagePool.Spec
func (client *NVCClient) ConstructRegionalNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID,
	zone string, availableZones []string, serviceLevel ServiceLevel) (storagePool *quarkv1alpha1.NetAppStoragePool,
	err error) {
	// Validate availableZones and zone
	zonesOk, err := client.CheckAvailableZones(serviceLevel, availableZones)
	if !zonesOk {
		return nil, err
	}

	storagePool, err = client.ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel)
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}

	// Set Zone and availableZones
	storagePool.Spec.Zone = zone
	storagePool.Spec.AvailableZones = availableZones
	return storagePool, err
}

// Constructs NetAppStoragePool definition.
// ownerID: OwnerID of the StoragePool.
// dispalyName: Name of the StoragePool.
// sizeString: StoragePool Size.
// storagePoolUUID: Unique ID of storagePool in k8s cluster.
// serviceLevel: Standard for alpha.
func (client *NVCClient) ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID string, serviceLevel ServiceLevel) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	defer func() {
		if e := recover(); e != nil {
			err = fmt.Errorf("invalid value for size")
		}
	}()

	// Validate service level
	switch serviceLevel {
	case ServiceLevelStandard, ServiceLevelRegionalStandard:
	default:
		return nil, fmt.Errorf("invalid value for service level")
	}

	// validate UUID
	_, err = uuid.FromString(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for UUID")
	}
	// Validate size
	size, err := resource.ParseQuantity(sizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	err = client.validateStoragePoolSize(size)
	if err != nil {
		return nil, err
	}

	// Validate ownerID
	if ownerID == "" {
		return nil, fmt.Errorf("storage pool OwnerID required")
	}

	storagePool = &quarkv1alpha1.NetAppStoragePool{
		metav1.TypeMeta{},
		metav1.ObjectMeta{
			Name: storagePoolUUID,
			Labels: map[string]string{
				quarkconfig.StoragePoolLabel: storagePoolUUID,
				quarkconfig.OwnerIDLabel:     ownerID,
			},
		},
		quarkv1alpha1.NetAppStoragePoolSpec{
			Size:         size,
			DisplayName:  displayName,
			ServiceLevel: string(serviceLevel),
		},
		quarkv1alpha1.NetAppStoragePoolStatus{},
	}
	return storagePool, err
}

func (client *NVCClient) NetAppStoragePoolCreate(storagePool *quarkv1alpha1.NetAppStoragePool, user, password string) (sp *quarkv1alpha1.NetAppStoragePool, err error) {

	var secretName string
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	// add quark version to create NetAppStoragePool only for "Production" & "TestVariant" configs
	switch strings.ToLower(quark.Spec.ControllerConfig.Variant) {
	case quarkconfig.ProductionVariant, quarkconfig.OnPremVariant:
		storagePool.Spec.QuarkVersion = quark.Status.Version
	}
	if storagePool.Spec.SMBServerConfiguration.Netbios != "" {
		secretName = util.GetVolCreateADSecretName(storagePool.Name)
		_, err = client.createOrUpdateADSecret(secretName, user, password)
		if err != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
			}).Error("Failed to create secret.")
			return nil, err
		}
	}

	sp, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Create(context.TODO(), storagePool, metav1.CreateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create storagePool.")
		return nil, err
	}

	return sp, nil
}
